body {
	background: #f3f3f3;
	{% if font.family %}
	font-family: {{ font.family }};
	{% endif %}
	{% if font.style %}
	font-style: {{ font.style }};
	{% endif %}
	{% if font.weight %}
	font-style: {{ font.weight }};
	{% endif %}
}

[contenteditable=true]:empty:before {
	content: attr(data-placeholder);
	display: block; /* For Firefox */
}

li[contenteditable=true]:hover, li[contenteditable=true].hovered, span[contenteditable=true]:hover, span[contenteditable=true].hovered {
	border: 1px solid #ff0000;
}

.rect:hover, a.hovered {
	box-shadow: inset 0 0 0 1px #ff0000;
}

li[contenteditable=true]{
	border: 1px dashed #000;
	width: 100%;
	padding: 2px;
	margin: 0 0 5px 0;
}

ul {
	list-style-type:none;
}

nav {
	background: #444;
	position: fixed;
	top: 0;
	left: 0;
	height: 100%;
	width: 10%;
	font-family: "Helvetica Neue", Arial, sans-serif;
}

nav ul {
	list-style: none;
	margin-right: 1em;
}

nav li {
	display : inline-block;
}

nav a {
	color: white;
	text-decoration: none;
}

nav a:hover {
	text-decoration: underline;	
}

.container {
	position: relative;
	margin-left: 15%;
	display: table;
	height: 100%;
	width: 85%;
}

.img_container {
	position: relative;
}

.column {
	display: table-cell;
	vertical-align: top;
	width: 50%;
	height: 100%;
	padding: 1rem;
}

#download_button {
	position: fixed;
	padding: 0;
	text-align: center;
	width: 10%;
	bottom: 50px;
}

.corrected {
	background-color: #73AD21;
}

